package and09.multiweatherapp.ui.home

import and09.multiweatherapp.R
import and09.multiweatherapp.weatherapi.FromLatLon
import and09.multiweatherapp.weatherapi.FromLocationName
import and09.multiweatherapp.weatherapi.OpenWeatherMapAPI
import and09.multiweatherapp.weatherapi.SpringWeatherAPI
import and09.multiweatherapp.weatherapi.WeatherAPI
import and09.multiweatherapp.weatherapi.WeatherStackAPI
import android.app.Application
import android.content.Context.LOCATION_SERVICE
import android.graphics.Bitmap
import android.graphics.BitmapFactory
import android.location.Location
import android.location.LocationManager
import android.util.Log
import android.widget.Spinner
import android.widget.Toast
import androidx.appcompat.app.AppCompatActivity
import androidx.core.content.ContextCompat.getSystemService
import androidx.lifecycle.AndroidViewModel
import androidx.lifecycle.LiveData
import androidx.lifecycle.MutableLiveData
import androidx.lifecycle.ViewModel
import androidx.preference.PreferenceManager
import kotlinx.coroutines.CoroutineScope
import kotlinx.coroutines.Dispatchers
import kotlinx.coroutines.delay
import kotlinx.coroutines.launch
import kotlinx.coroutines.withContext
import java.io.FileNotFoundException
import java.net.ConnectException
import java.net.URL
import java.net.UnknownHostException
import kotlin.reflect.KFunction
import kotlin.reflect.full.companionObject
import kotlin.reflect.full.companionObjectInstance
import kotlin.reflect.full.declaredFunctions
import kotlin.reflect.full.hasAnnotation

class HomeViewModel(application: Application) : AndroidViewModel(application) {
    private val _location: MutableLiveData<String> by lazy {
        MutableLiveData<String>()
    }
    private val _temperature: MutableLiveData<String> by lazy {
        MutableLiveData<String>()
    }
    private val _description: MutableLiveData<String> by lazy {
        MutableLiveData<String>()
    }
    private val _provider: MutableLiveData<String> by lazy {
        MutableLiveData<String>()
    }
    private val _iconBitmap: MutableLiveData<Bitmap> by lazy {
        MutableLiveData<Bitmap>()
    }
    val location: LiveData<String> = _location
    val temperature: LiveData<String> = _temperature
    val description: LiveData<String> = _description
    val provider: LiveData<String> = _provider
    val iconBitmap: LiveData<Bitmap> = _iconBitmap

    fun retrieveWeatherData() {
        CoroutineScope(Dispatchers.Main).launch() {
            var errorMessage : String = ""
            var weather : WeatherAPI? = null
            var bitmap : Bitmap? = null
            var lastLocation : Location? = null
            lateinit var func: KFunction<*>

            withContext(Dispatchers.IO) {
                val app = getApplication() as Application
                val prefs = PreferenceManager.getDefaultSharedPreferences(app)
                if (prefs.getBoolean("use_gps", false)) {
                    try {
                        val locationManager =
                            app.getSystemService(LOCATION_SERVICE) as LocationManager
                        lastLocation = locationManager.getLastKnownLocation("gps")
                    } catch (ex: SecurityException) {
                        Log.e(
                            javaClass.simpleName,
                            "Erforderliche Berechtigung $ex.toString() nicht erteilt"
                        )
                    }
                }
                val locationName = prefs.getString(app.getString(R.string.location_name), "Witzenhausen")
                val providerClassName = prefs.getString(app.getString(R.string.weather_provider), app.resources.getStringArray(R.array.class_names)[0])
                try {
                    val cls = Class.forName("and09.multiweatherapp.weatherapi.$providerClassName").kotlin
                    if (lastLocation == null) {
                        func = cls.companionObject?.declaredFunctions?.find{ it.hasAnnotation<FromLocationName>()}!!
                    } else {
                        func = cls.companionObject?.declaredFunctions?.find{ it.hasAnnotation<FromLatLon>()}!!
                    }
                    val compObj = cls.companionObjectInstance
                    if (providerClassName.equals(app.resources.getStringArray(R.array.class_names)[2])) {
                        val serverName = prefs.getString("server_name", "localhost")
                        if (!serverName.equals("localhost")) {
                            val func2 = cls.companionObject?.declaredFunctions?.find{ it.name  == "setServerAddress"}
                            func2?.call(compObj, serverName)
                        }
                    }
                    if (lastLocation == null) {
                        weather = func?.call(compObj, locationName) as WeatherAPI
                    } else {
                        weather = func?.call(compObj, lastLocation!!.latitude, lastLocation!!.longitude) as WeatherAPI
                    }

                    Log.d(javaClass.simpleName, "Temp: ${weather?.temperature}")
                    Log.d(javaClass.simpleName, "Description: ${weather?.description}")
                    Log.d(javaClass.simpleName, "Icon-URL: ${weather?.iconUrl}")
                    Log.d(javaClass.simpleName, "Provider: ${weather?.providerUrl}")
                    val iconUrl = URL(weather?.iconUrl) // java.net!
                    val inputStream = iconUrl.openStream()
                    bitmap = BitmapFactory.decodeStream(inputStream)
                    inputStream.close()
                } catch (ex: Exception) {
                    Log.e(javaClass.simpleName, ex.toString())
                    errorMessage = when(ex.cause) {
                        is UnknownHostException -> "Keine Internetverbindung. Bitte überprüfen Sie Ihre Netzwerkeinstellungen"
                        is ConnectException -> "Netzwerkdienst antwortet nicht. Bitte schalten Sie auf einen anderen Dienst um"
                        is FileNotFoundException -> "Es wurden keine Daten zum gewählten Standort zurückgeliefert"
                        else -> "Unbekannter Fehler"
                    }
                }
            }
            if (weather != null) updateValues(weather, bitmap)
            else Toast.makeText(getApplication(), errorMessage, Toast.LENGTH_LONG).show()
        }
    }

    fun updateValues(weather: WeatherAPI?, bitmap: Bitmap?) {
        _location.value = weather?.location
        _temperature.value = "${weather?.temperature} °C"
        _description.value = weather?.description
        _provider.value = weather?.providerUrl
        _iconBitmap.value = bitmap
    }
}